import dash_bootstrap_components as dbc
from dash import Input, Output, State, callback, dcc, html, register_page
from openspace.bodies.celestial import Earth
from openspace.coordinates.states import GCRF, HCW, StateConvert
from openspace.math.functions import EquationsOfMotion
from openspace.math.linalg import Vector3D, Vector6D
from openspace.time import Epoch

register_page(__name__, path="/", title="Dashboard", name="Dashboard")
config_accordion = dbc.Accordion(
    [
        dbc.AccordionItem(
            [
                dbc.Label("Target Epoch"),
                dbc.Input(
                    id="target-epoch-input",
                    persistence=True,
                    type="text",
                    className="epoch-input",
                    value="2023-01-30 12:00:00",
                ),
                dbc.FormText(
                    "This epoch represents the time at which the target state is valid.  Inputs are in Terrestrial \
                    Dynamic Time (TDT).  "
                ),
                dbc.Label("Format:  YYYY-MM-DD hh:mm:ss"),
            ],
            title="Epochs",
        ),
        dbc.AccordionItem(
            [
                dbc.Label("Target State"),
                dbc.InputGroup(
                    [
                        dbc.Input(
                            id="target-input-x",
                            type="number",
                            value=42164,
                            persistence=True,
                            className="vector6d-field",
                        ),
                        dbc.Input(
                            id="target-input-y", type="number", value=0, persistence=True, className="vector6d-field"
                        ),
                        dbc.Input(
                            id="target-input-z", type="number", value=0, persistence=True, className="vector6d-field"
                        ),
                        dbc.Input(
                            id="target-input-vx", type="number", value=0, persistence=True, className="vector6d-field"
                        ),
                        dbc.Input(
                            id="target-input-vy",
                            type="number",
                            value=3.074,
                            persistence=True,
                            className="vector6d-field",
                        ),
                        dbc.Input(
                            id="target-input-vz", type="number", value=0, persistence=True, className="vector6d-field"
                        ),
                    ]
                ),
                dbc.FormText(
                    " This state will act as the origin when defining the chase vehicle's \
                    state.  Inputs are in the GCRF frame."
                ),
                html.Br(),
                dbc.Label("Chase State", style={"margin-top": "5%"}),
                dbc.Row(
                    [
                        dbc.Col(dbc.FormText(id="chase-text-x")),
                        dbc.Col(dbc.FormText(id="chase-text-y")),
                        dbc.Col(dbc.FormText(id="chase-text-z")),
                        dbc.Col(dbc.FormText(id="chase-text-vx")),
                        dbc.Col(dbc.FormText(id="chase-text-vy")),
                        dbc.Col(dbc.FormText(id="chase-text-vz")),
                    ]
                ),
                dbc.FormText(
                    "This represents the chase vehicle's state in the inertial frame.  Direct edits cannot be made \
                    here.  Instead, use the relative motion tab."
                ),
                html.Br(),
                dbc.Label("Format:", style={"margin-top": "5%"}),
                dbc.Row(
                    [
                        dbc.Col(dcc.Markdown(r"$x (km)$", mathjax=True)),
                        dbc.Col(dcc.Markdown(r"$y (km)$", mathjax=True)),
                        dbc.Col(dcc.Markdown(r"$z (km)$", mathjax=True)),
                        dbc.Col(dcc.Markdown(r"$\dot x (\frac{km}{s})$", mathjax=True)),
                        dbc.Col(dcc.Markdown(r"$\dot y (\frac{km}{s})$", mathjax=True)),
                        dbc.Col(dcc.Markdown(r"$\dot z (\frac{km}{s})$", mathjax=True)),
                    ]
                ),
            ],
            title="States",
        ),
    ],
)
content_column = dbc.Col(
    [
        dbc.FormText(
            "Openspace is an opensource solution to training and education of space operations. \
            The backend library can be installed from a command line using 'pip install openspace' or by downloading \
            the source code from https://github.com/brandon-sexton/openspace.  For information on backend \
            capabilities, visit https://www.openspace-docs.com.  For questions, function requests, or to report \
            an issue, please e-mail brandon.sexton.1@outlook.com.  Otherwise, select a widget link from the \
            toolbar to get started.",
        ),
        config_accordion,
    ],
    className="content-container",
)

nav_column = dbc.Col(
    dbc.Nav(
        [
            dbc.NavItem(dbc.NavLink("Dashboard", href="/")),
            dbc.NavItem(dbc.NavLink("Relative Motion", href="/cw")),
            dbc.NavItem(dbc.NavLink("Inertial View", href="/inertial")),
            dbc.NavItem(dbc.NavLink("State Estimation", href="/od")),
            dbc.NavItem(dbc.NavLink("Documentation", href="https://www.openspace-docs.com")),
            dbc.NavItem(dbc.NavLink("Source Code", href="https://github.com/brandon-sexton/")),
        ],
        vertical="sm",
        pills=True,
    ),
    width="auto",
)

layout = dbc.Container([html.Br(), dbc.Row([nav_column, content_column])])


@callback(
    [
        Output("target-epoch", "data"),
        Output("target-epoch-input", "invalid"),
    ],
    Input("target-epoch-input", "value"),
    State("target-epoch", "data"),
)
def update_target_epoch(ep_str, tgt_ep):

    ep = tgt_ep
    invalid = False
    date_time = ep_str.split(" ")
    if len(date_time) < 2:
        invalid = True
    else:
        date_list = date_time[0].split("-")
        time_list = date_time[1].split(":")
        if len(date_list) < 3 or len(time_list) < 3:
            invalid = True
        else:
            year = float(date_list[0])
            month = float(date_list[1])
            day = float(date_list[2])

            hrs = float(time_list[0])
            mins = float(time_list[1])
            secs = float(time_list[2])
            if hrs >= 24 or hrs < 0:
                invalid = True
            elif mins >= 60 or mins < 0:
                invalid = True
            elif secs >= 60 or secs < 0:
                invalid = True
            elif year < 1858:
                invalid = True
            elif month < 1 or month > 12:
                invalid = True
            elif day < 0:
                invalid = True
            elif (month == 9 or month == 4 or month == 6 or month == 10) and day > 30:
                invalid = True
            elif month == 2 and day > 29:
                invalid = True

    if not invalid:
        ep = Epoch.from_gregorian(year, month, day, hrs, mins, secs)

    return invalid, ep


@callback(
    [
        Output("chase-text-x", "children"),
        Output("chase-text-y", "children"),
        Output("chase-text-z", "children"),
        Output("chase-text-vx", "children"),
        Output("chase-text-vy", "children"),
        Output("chase-text-vz", "children"),
        Output("sma", "data"),
        Output("target-x", "data"),
        Output("target-y", "data"),
        Output("target-z", "data"),
        Output("target-vx", "data"),
        Output("target-vy", "data"),
        Output("target-vz", "data"),
    ],
    [
        Input("target-epoch", "data"),
        Input("target-input-x", "value"),
        Input("target-input-y", "value"),
        Input("target-input-z", "value"),
        Input("target-input-vx", "value"),
        Input("target-input-vy", "value"),
        Input("target-input-vz", "value"),
    ],
    [
        State("r-pos", "data"),
        State("i-pos", "data"),
        State("c-pos", "data"),
        State("r-vel", "data"),
        State("i-vel", "data"),
        State("c-vel", "data"),
    ],
)
def update_chase(tgt_ep, x, y, z, vx, vy, vz, r, i, c, vr, vi, vc):

    ep: Epoch = Epoch(tgt_ep)
    tgt: GCRF = GCRF(ep, Vector3D(x, y, z), Vector3D(vx, vy, vz))
    hill: HCW = HCW.from_state_vector(Vector6D(r, i, c, vr, vi, vc))
    chase: GCRF = StateConvert.hcw.to_gcrf(hill, tgt)
    sma: float = EquationsOfMotion.A.from_mu_r_v(Earth.MU, tgt.position.magnitude(), tgt.velocity.magnitude())
    cx = "%.6f" % chase.position.x
    cy = "%.6f" % chase.position.y
    cz = "%.6f" % chase.position.z
    cvx = "%.6f" % chase.velocity.x
    cvy = "%.6f" % chase.velocity.y
    cvz = "%.6f" % chase.velocity.z
    return cx, cy, cz, cvx, cvy, cvz, sma, x, y, z, vx, vy, vz



import dash_bootstrap_components as dbc
import plotly.graph_objects as go
from dash import callback, dcc, html, register_page
from dash.dependencies import Input, Output, State
from openspace.bodies.artificial import Spacecraft
from openspace.coordinates.states import GCRF, HCW, StateConvert
from openspace.math.linalg import Vector3D, Vector6D
from openspace.time import Epoch

register_page(__name__, title="Inertial", name="Inertial")

figure = {
    "data": [
        {"type": "scatter3d", "mode": "markers", "marker": {"color": "blue"}, "name": "Earth"},
        {"type": "scatter3d", "mode": "lines", "line": {"color": "darkmagenta"}, "name": "Target"},
        {"type": "scatter3d", "mode": "lines", "line": {"color": "darkcyan"}, "name": "Chase"},
    ],
    "layout": go.Layout(
        autosize=True,
        uirevision="constant",
        template="plotly_dark",
    ),
}

nav_column = dbc.Col(
    dbc.Nav(
        [
            dbc.NavItem(dbc.NavLink("Dashboard", href="/")),
            dbc.NavItem(dbc.NavLink("Relative Motion", href="/cw")),
            dbc.NavItem(dbc.NavLink("Inertial View", href="/inertial")),
            dbc.NavItem(dbc.NavLink("State Estimation", href="/od")),
            dbc.NavItem(dbc.NavLink("Documentation", href="https://www.openspace-docs.com")),
            dbc.NavItem(dbc.NavLink("Source Code", href="https://github.com/brandon-sexton/")),
        ],
        vertical="sm",
        pills=True,
    ),
    width="auto",
)

layout = dbc.Container(
    [
        html.Br(),
        dbc.Row(
            [
                nav_column,
                dbc.Col(
                    [
                        html.Div(
                            [
                                html.H2("Inertial Visualization"),
                                dbc.FormText(
                                    "This view shows how the target and chase state look independent of each other."
                                ),
                            ]
                        ),
                        dcc.Graph(
                            id="eci-plot", responsive=True, figure=figure, style={"width": "100%", "height": "80%"}
                        ),
                    ],
                    className="content-container",
                ),
            ],
            style={"height": "100vh"},
        ),
    ]
)


@callback(
    Output("eci-plot", "figure"),
    [
        Input("target-x", "data"),
        Input("target-y", "data"),
        Input("target-z", "data"),
        Input("target-vx", "data"),
        Input("target-vy", "data"),
        Input("target-vz", "data"),
        Input("target-epoch", "data"),
        Input("r-pos", "data"),
        Input("i-pos", "data"),
        Input("c-pos", "data"),
        Input("r-vel", "data"),
        Input("i-vel", "data"),
        Input("c-vel", "data"),
    ],
    State("eci-plot", "figure"),
)
def update_plot(x, y, z, vx, vy, vz, tgt_ep, r, i, c, vr, vi, vc, figure):
    ep = Epoch(tgt_ep)
    tgt = Spacecraft(GCRF(ep, Vector3D(x, y, z), Vector3D(vx, vy, vz)))
    chase = Spacecraft(
        StateConvert.hcw.to_gcrf(HCW.from_state_vector(Vector6D(r, i, c, vr, vi, vc)), tgt.current_state())
    )
    end_ep = ep.plus_days(1)

    tx, ty, tz, cx, cy, cz = [], [], [], [], [], []
    while tgt.current_epoch().value < end_ep.value:
        tgt.step()
        chase.step()
        tx.append(tgt.position().x)
        ty.append(tgt.position().y)
        tz.append(tgt.position().z)
        cx.append(chase.position().x)
        cy.append(chase.position().y)
        cz.append(chase.position().z)

    figure = {
        "data": [
            {
                "x": [0],
                "y": [0],
                "z": [0],
                "type": "scatter3d",
                "mode": "markers",
                "marker": {"color": "blue"},
                "name": "Earth",
            },
            {
                "x": tx,
                "y": ty,
                "z": tz,
                "type": "scatter3d",
                "mode": "lines",
                "line": {"color": "darkmagenta"},
                "name": "Target",
            },
            {
                "x": cx,
                "y": cy,
                "z": cz,
                "type": "scatter3d",
                "mode": "lines",
                "line": {"color": "darkcyan"},
                "name": "Chase",
            },
        ],
        "layout": go.Layout(
            autosize=True,
            uirevision="constant",
            template="plotly_dark",
        ),
    }

    return figure



import dash_bootstrap_components as dbc
import plotly.graph_objects as go
from dash import callback, dcc, html, register_page
from dash.dependencies import Input, Output, State
from openspace.coordinates.states import HCW
from openspace.math.constants import BASE_IN_KILO, SECONDS_IN_DAY
from openspace.math.linalg import Vector6D
from openspace.propagators.relative import Hill

register_page(__name__, title="Relative Motion", name="Relative Motion")

figure = {
    "data": [
        {"type": "scatter3d", "mode": "lines", "line": {"color": "darkcyan"}, "name": "Chase"},
        {"type": "scatter3d", "mode": "markers", "marker": {"color": "darkmagenta"}, "name": "Target"},
    ],
    "layout": go.Layout(
        uirevision="constant",
        template="plotly_dark",
        scene={
            "xaxis": {"range": [-300, 300], "title": "radial"},
            "yaxis": {"range": [-300, 300], "title": "in-track"},
            "zaxis": {"range": [-300, 300], "title": "cross-track"},
        },
    ),
}

nav_column = dbc.Col(
    dbc.Nav(
        [
            dbc.NavItem(dbc.NavLink("Dashboard", href="/")),
            dbc.NavItem(dbc.NavLink("Relative Motion", href="/cw")),
            dbc.NavItem(dbc.NavLink("Inertial View", href="/inertial")),
            dbc.NavItem(dbc.NavLink("State Estimation", href="/od")),
            dbc.NavItem(dbc.NavLink("Documentation", href="https://www.openspace-docs.com")),
            dbc.NavItem(dbc.NavLink("Source Code", href="https://github.com/brandon-sexton/")),
        ],
        vertical="sm",
        pills=True,
    ),
    width="auto",
)

content_column = dbc.Col(
    [
        html.Div(
            [
                html.H2("3-D Relative Motion Visualization and Planning"),
                dbc.FormText(
                    "These values are relative to the target state defined in the dashboard tab.  Input various values \
                    to see how state estimation ability changes."
                ),
            ]
        ),
        dbc.InputGroup(
            [
                dbc.Input(id="r-pos-input", type="number", value=-5, persistence=True, style={"width": "16.666%"}),
                dbc.Input(id="i-pos-input", type="number", value=0, persistence=True, style={"width": "16.666%"}),
                dbc.Input(id="c-pos-input", type="number", value=0, persistence=True, style={"width": "16.666%"}),
                dbc.Input(id="r-vel-input", type="number", value=0, persistence=True, style={"width": "16.666%"}),
                dbc.Input(id="i-vel-input", type="number", value=0, persistence=True, style={"width": "16.666%"}),
                dbc.Input(id="c-vel-input", type="number", value=1, persistence=True, style={"width": "16.666%"}),
            ]
        ),
        dbc.Row(
            [
                dbc.Col(dcc.Markdown(r"$x (km)$", mathjax=True)),
                dbc.Col(dcc.Markdown(r"$y (km)$", mathjax=True)),
                dbc.Col(dcc.Markdown(r"$z (km)$", mathjax=True)),
                dbc.Col(dcc.Markdown(r"$\dot x (\frac{km}{s})$", mathjax=True)),
                dbc.Col(dcc.Markdown(r"$\dot y (\frac{km}{s})$", mathjax=True)),
                dbc.Col(dcc.Markdown(r"$\dot z (\frac{km}{s})$", mathjax=True)),
            ]
        ),
        dcc.Graph(id="rel-plot", responsive=True, figure=figure, style={"width": "100%", "height": "80%"}),
    ],
    className="content-container",
)
layout = dbc.Container(
    [
        html.Br(),
        dbc.Row([nav_column, content_column], style={"height": "100vh"}),
    ]
)


@callback(
    Output("rel-plot", "figure"),
    [
        Input("r-pos-input", "value"),
        Input("i-pos-input", "value"),
        Input("c-pos-input", "value"),
        Input("r-vel-input", "value"),
        Input("i-vel-input", "value"),
        Input("c-vel-input", "value"),
    ],
    [
        State("sma", "data"),
        State("rel-plot", "figure"),
    ],
)
def update_plot(r, i, c, vr, vi, vc, sma, figure):
    m_to_km = 1 / BASE_IN_KILO
    input_hcw: HCW = HCW.from_state_vector(Vector6D(r, i, c, vr * m_to_km, vi * m_to_km, vc * m_to_km))
    prop = Hill(input_hcw, sma)
    total_time = 0
    dt = prop.step_size
    r, i, c = [], [], []
    prop.step_by_seconds(-SECONDS_IN_DAY * 0.5)

    while total_time < SECONDS_IN_DAY:
        prop.step()
        r.append(prop.state.position.x)
        i.append(prop.state.position.y)
        c.append(prop.state.position.z)
        total_time += dt

    figure = {
        "data": [
            {
                "x": r,
                "y": i,
                "z": c,
                "type": "scatter3d",
                "mode": "lines",
                "line": {"color": "darkcyan"},
                "name": "Chase",
            },
            {
                "x": [0],
                "y": [0],
                "z": [0],
                "type": "scatter3d",
                "mode": "markers",
                "marker": {"color": "darkmagenta"},
                "name": "Target",
            },
        ],
        "layout": go.Layout(
            autosize=True,
            uirevision="constant",
            template="plotly_dark",
            scene={
                "xaxis": {"range": [-300, 300], "title": "radial"},
                "yaxis": {"range": [-300, 300], "title": "in-track"},
                "zaxis": {"range": [-300, 300], "title": "cross-track"},
            },
        ),
    }

    return figure


@callback(
    [
        Output("r-pos", "data"),
        Output("i-pos", "data"),
        Output("c-pos", "data"),
    ],
    [
        Input("r-pos-input", "value"),
        Input("i-pos-input", "value"),
        Input("c-pos-input", "value"),
    ],
)
def update_position(r, i, c):
    return r, i, c


@callback(
    [
        Output("r-vel", "data"),
        Output("i-vel", "data"),
        Output("c-vel", "data"),
    ],
    [
        Input("r-vel-input", "value"),
        Input("i-vel-input", "value"),
        Input("c-vel-input", "value"),
    ],
)
def update_velocity(r, i, c):
    return r / BASE_IN_KILO, i / BASE_IN_KILO, c / BASE_IN_KILO



import webbrowser
from threading import Timer

import dash
import dash_bootstrap_components as dbc
import flask
from dash import Dash, dcc, html

app = Dash(
    __name__,
    use_pages=True,
    external_stylesheets=[dbc.themes.BOOTSTRAP, "\\assets\\css\\custom-style.css"],
    server=flask.Flask(__name__),
)

app.layout = dbc.Container(
    [
        html.Br(),
        dbc.Row(
            html.Img(className="header-img", src=dash.get_asset_url("img/openspace-header.png")),
        ),
        dbc.Row(
            dash.page_container,
        ),
        dcc.Store(id="r-pos", storage_type="session", data=0),
        dcc.Store(id="i-pos", storage_type="session", data=0),
        dcc.Store(id="c-pos", storage_type="session", data=0),
        dcc.Store(id="r-vel", storage_type="session", data=0),
        dcc.Store(id="i-vel", storage_type="session", data=0),
        dcc.Store(id="c-vel", storage_type="session", data=0),
        dcc.Store(id="target-epoch-input", storage_type="session", data=0),
        dcc.Store(id="target-epoch", storage_type="session", data=0),
        dcc.Store(id="year-input", storage_type="session", data=0),
        dcc.Store(id="month-input", storage_type="session", data=0),
        dcc.Store(id="day-input", storage_type="session", data=0),
        dcc.Store(id="hour-input", storage_type="session", data=0),
        dcc.Store(id="minute-input", storage_type="session", data=0),
        dcc.Store(id="second-input", storage_type="session", data=0),
        dcc.Store(id="target-x", storage_type="session", data=0),
        dcc.Store(id="target-y", storage_type="session", data=0),
        dcc.Store(id="target-z", storage_type="session", data=0),
        dcc.Store(id="target-vx", storage_type="session", data=0),
        dcc.Store(id="target-vy", storage_type="session", data=0),
        dcc.Store(id="target-vz", storage_type="session", data=0),
        dcc.Store(id="sma", storage_type="session", data=0),
    ],
)


def run():
    host = "localhost"
    port = 8888
    url = f"http://{host}:{port}"
    Timer(10, webbrowser.open_new(url))

    # run app
    app.run(host=host, port=port, debug=False)


if __name__ == "__main__":
    run()



import dash_bootstrap_components as dbc
import plotly.graph_objects as go
from dash import callback, dcc, html, register_page
from dash.dependencies import Input, Output, State
from openspace.bodies.artificial import Spacecraft
from openspace.coordinates.states import GCRF, HCW, StateConvert
from openspace.math.linalg import Vector3D, Vector6D
from openspace.time import Epoch

register_page(__name__, title="State Estimation", name="State Estimation")

nav_column = dbc.Col(
    dbc.Nav(
        [
            dbc.NavItem(dbc.NavLink("Dashboard", href="/")),
            dbc.NavItem(dbc.NavLink("Relative Motion", href="/cw")),
            dbc.NavItem(dbc.NavLink("Inertial View", href="/inertial")),
            dbc.NavItem(dbc.NavLink("State Estimation", href="/od")),
            dbc.NavItem(dbc.NavLink("Documentation", href="https://www.openspace-docs.com")),
            dbc.NavItem(dbc.NavLink("Source Code", href="https://github.com/brandon-sexton/")),
        ],
        vertical="sm",
        pills=True,
    ),
    width="auto",
)

figure = {
    "data": [
        {"type": "scatter3d", "mode": "markers", "marker": {"color": "darkmagenta"}, "name": "Target"},
        {"type": "scatter3d", "mode": "lines", "line": {"color": "darkcyan"}, "name": "Truth"},
        {"type": "scatter3d", "mode": "lines", "line": {"color": "firebrick"}, "name": "Observed"},
    ],
    "layout": go.Layout(
        autosize=True,
        uirevision="constant",
        template="plotly_dark",
    ),
}
layout = dbc.Container(
    [
        html.Br(),
        dbc.Row(
            [
                nav_column,
                dbc.Col(
                    [
                        html.Div(
                            [
                                html.H2("Filter Performance"),
                                dbc.FormText(
                                    "This page shows how well the chase vehicle is able to estimate the state of the \
                                    target given the pre-defined relative state. Ideally, the observed state would \
                                    always be on top of truth.  The greater the difference in the observed and \
                                    truth state, the worse the profile is for state estimation."
                                ),
                            ]
                        ),
                        dcc.Graph(
                            id="od-plot", responsive=True, style={"width": "100%", "height": "80%"}, figure=figure
                        ),
                    ],
                    className="content-container",
                ),
            ],
            style={"height": "100vh"},
        ),
    ]
)


@callback(
    Output("od-plot", "figure"),
    [
        Input("target-x", "data"),
        Input("target-y", "data"),
        Input("target-z", "data"),
        Input("target-vx", "data"),
        Input("target-vy", "data"),
        Input("target-vz", "data"),
        Input("target-epoch", "data"),
        Input("r-pos", "data"),
        Input("i-pos", "data"),
        Input("c-pos", "data"),
        Input("r-vel", "data"),
        Input("i-vel", "data"),
        Input("c-vel", "data"),
    ],
    State("od-plot", "figure"),
)
def update_plot(x, y, z, vx, vy, vz, tgt_ep, r, i, c, vr, vi, vc, figure):
    ep = Epoch(tgt_ep)
    tgt = Spacecraft(GCRF(ep, Vector3D(x, y, z), Vector3D(vx, vy, vz)))
    chase = Spacecraft(
        StateConvert.hcw.to_gcrf(HCW.from_state_vector(Vector6D(r, i, c, vr, vi, vc)), tgt.current_state())
    )
    tgt.step_to_epoch(ep.plus_days(-0.5))
    chase.step_to_epoch(ep.plus_days(-0.5))
    end_ep = ep.plus_days(0.5)

    seed = Spacecraft(GCRF(ep, Vector3D(x + 0.5, y + 0.5, z + 0.5), Vector3D(vx, vy, vz)))
    seed.step_to_epoch(ep.plus_days(-0.5))
    chase.acquire(seed)
    tx, ty, tz, cx, cy, cz = [], [], [], [], [], []
    while tgt.current_epoch().value < end_ep.value:
        tgt.step()
        chase.step()
        chase.process_wfov(tgt)

        rel_truth = tgt.hill_position(chase)
        tx.append(-chase.filter.propagator.state.position.x)
        ty.append(-chase.filter.propagator.state.position.y)
        tz.append(-chase.filter.propagator.state.position.z)
        cx.append(rel_truth.x)
        cy.append(rel_truth.y)
        cz.append(rel_truth.z)

    figure = {
        "data": [
            {
                "x": [0],
                "y": [0],
                "z": [0],
                "type": "scatter3d",
                "mode": "markers",
                "marker": {"color": "darkmagenta"},
                "name": "Target",
            },
            {
                "x": cx,
                "y": cy,
                "z": cz,
                "type": "scatter3d",
                "mode": "lines",
                "line": {"color": "darkcyan"},
                "name": "Truth",
            },
            {
                "x": tx,
                "y": ty,
                "z": tz,
                "type": "scatter3d",
                "mode": "lines",
                "line": {"color": "firebrick"},
                "name": "Observed",
            },
        ],
        "layout": go.Layout(
            autosize=True,
            uirevision="constant",
            template="plotly_dark",
        ),
    }

    return figure
